#include <limits>
#include <msp/core/maputils.h>
#include "binarywriter.h"
#include "binfloat.h"
#include "output.h"
#include "statement.h"

using namespace std;

namespace Msp {
namespace DataFile {

BinaryWriter::BinaryWriter(Output &o):
	WriterMode(o),
	next_kwd_id(1),
	next_str_id(1),
	float_precision(32)
{
	dict[StatementKey("__kwd", "iss")] = -1;
	dict[StatementKey("__str", "is")] = -2;
	dict[StatementKey("__flt", "i")] = -3;
}

void BinaryWriter::set_float_precision(unsigned fp)
{
	if(fp<16 || fp>64 || fp%8)
		throw invalid_argument("BinaryWriter::set_float_precision");
	float_precision = fp;
	Statement fst;
	fst.keyword = "__flt";
	fst.args.push_back(float_precision);
	write_(fst);
}

void BinaryWriter::write(const Statement &st)
{
	collect_keywords(st);
	write_(st);
}

void BinaryWriter::write_(const Statement &st)
{
	int id = get_item(dict, StatementKey(st.keyword, st.get_signature()));

	write_int(id);
	for(const Value &a: st.args)
		switch(a.get_signature())
		{
		case IntType::signature:    write_int   (a.get<IntType::Store>()); break;
		case StringType::signature: write_string(a.get<StringType::Store>()); break;
		case BoolType::signature:   write_int   (a.get<BoolType::Store>()); break;
		case FloatType::signature:  write_float (a.get<FloatType::Store>()); break;
		case SymbolType::signature: write_symbol(a.get<SymbolType::Store>()); break;
		}

	write_int(st.sub.size());
	for(const Statement &s: st.sub)
		write(s);
}

void BinaryWriter::collect_keywords(const Statement &st)
{
	StatementKey key(st.keyword, st.get_signature());

	if(!dict.count(key))
	{
		Statement kst;
		kst.keyword = "__kwd";
		kst.args.push_back(next_kwd_id);
		kst.args.push_back(key.keyword);
		kst.args.push_back(key.signature);
		write_(kst);

		dict[key] = next_kwd_id++;
	}

	for(const Value &a: st.args)
	{
		char sig = a.get_signature();
		string str;
		if(sig==SymbolType::signature)
			str = a.get<Symbol>().name;
		else if(sig==StringType::signature)
		{
			str = a.get<string>();
			if(str.size()>32)
				continue;
		}
		else
			continue;

		if(strings.count(str))
			continue;

		Statement sst;
		sst.keyword = "__str";
		sst.args.push_back(next_str_id);
		sst.args.push_back(str);
		write_(sst);

		strings[str] = next_str_id++;
	}

	for(const Statement &s: st.sub)
		collect_keywords(s);
}

void BinaryWriter::write_int(IntType::Store n)
{
	unsigned i = sizeof(IntType::Store)-1;

	if(n>=0)
		for(; (i>0 && (n>>(i*7-1))==0); --i) ;
	else
		for(; (i>0 && (n>>(i*7-1))==-1); --i) ;

	for(++i; i--;)
		out.put((n>>(i*7) & 0x7F) | (i?0x80:0));
}

void BinaryWriter::write_string(const StringType::Store &s)
{
	StringMap::const_iterator i = strings.find(s);
	if(i!=strings.end())
		write_int(-static_cast<int>(i->second));
	else
	{
		write_int(s.size());
		out.write(s);
	}
}

void BinaryWriter::write_float(FloatType::Store f)
{
	BinFloat bf;

	if(numeric_limits<FloatType::Store>::is_iec559)
		bf = BinFloat::explode_iec559(f);
	else
	{
		/* The structure of the float is unknown, so we must use arithmetic to
		reduce it to components. */
		bf.sign = f<0;
		bf.exponent = 0;
		bf.mantissa = 0;

		if(f<0)
			f = -f;
		if(!(f+f>f))
			bf.infinity = true;
		else if(f!=0)
		{
			for(; f<1; f*=2)
				--bf.exponent;
			for(; f>=2; f/=2)
				++bf.exponent;
			for(unsigned i=0; i<64; ++i)
			{
				bf.mantissa <<= 1;
				if(f>=1)
				{
					bf.mantissa |= 1;
					f -= 1;
				}
				f *= 2;
			}
		}
	}

	UInt64 encoded = bf.compose(float_precision);
	for(unsigned i=float_precision/8; i--; )
		out.put((encoded>>(i*8))&0xFF);
}

void BinaryWriter::write_symbol(const SymbolType::Store &s)
{
	write_int(get_item(strings, s.name));
}

} // namespace DataFile
} // namespace Msp
